{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "ef0fcfdd",
   "metadata": {},
   "source": [
    "# Week IX - The Ising model"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "2ba562f9",
   "metadata": {},
   "source": [
    "Today, we will discuss a possible implementation of the **2D Ising model on a square lattice**. In particular, we will consider a system in equilibrium with a thermal bath, such that the temperature is constant. Temporal evolution will be simulated through the **spin flip dynamics** in periodic boundary conditions (PBC).\n",
    "\n",
    "We will study the model by monitoring the following quantities:\n",
    "* spin magnetisation \n",
    "    $$\\frac{M}{N}  = \\frac{1}{N}\\sum_i s_i$$\n",
    "* energy \n",
    "    $$\\frac{E}{N} = \\frac{1}{N}\\sum_{NN(i,j)} s_i$$\n",
    "* heat capacity\n",
    "    $$c = \\frac{<E^2>-<E>^2}{k_BT^2N}$$\n",
    "* magnetic susceptibility\n",
    "    $$\\chi = \\frac{<M^2>-<M>^2}{k_BTN}$$\n",
    "\n",
    "In the following we will set $k_B=1$ and $J=1$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "cc2f0ade",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "from matplotlib.colors import ListedColormap\n",
    "\n",
    "class Ising:\n",
    "    '''\n",
    "    This class implements \n",
    "    the 2D Ising model\n",
    "    on a square lattice in PBCs\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,L,T,N_equi,N_mcs,sample_equi=False,data_filename='data.txt'):\n",
    "        'Initialise the Ising model'\n",
    "        self.L = L\n",
    "        self.N = L**2\n",
    "        self.T = T\n",
    "        self.M = None\n",
    "        self.M_N = None\n",
    "        self.E = None\n",
    "        self.E_N = None\n",
    "        self.E_N_av = None\n",
    "        self.M_N_av = None\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.spin = np.zeros((L,L),dtype=np.int32)\n",
    "        self.w_dict = {}\n",
    "        self.acc = 0\n",
    "        self.counter = 0\n",
    "        self.N_equi = N_equi\n",
    "        self.N_mcs = N_mcs\n",
    "        self.data = []\n",
    "        self.ar = None\n",
    "        self.var = None\n",
    "        self.sample_equi = sample_equi\n",
    "        self.c_heat = None\n",
    "        self.chi_mag = None\n",
    "        self.autocorr = []\n",
    "        self.n_autocorr = 10\n",
    "        self.data_filename = data_filename \n",
    "    \n",
    "    def make_pm(self,x):\n",
    "        '''\n",
    "        Convert random uniform array\n",
    "        into array of -1 and +1\n",
    "        '''\n",
    "        def tmp(x):\n",
    "            if x<0.5:\n",
    "                return 1\n",
    "            else:\n",
    "                return -1\n",
    "        f = np.frompyfunc(tmp, 1, 1)\n",
    "        return f(x)\n",
    "    \n",
    "    def get_up_right(self,i,j):\n",
    "        '''\n",
    "        PBCs up/right\n",
    "        '''\n",
    "        if j==(self.L-1):\n",
    "            up = 0\n",
    "        else:\n",
    "            up = j+1\n",
    "        if i==(self.L-1):\n",
    "            right = 0\n",
    "        else:\n",
    "            right = i+1\n",
    "            \n",
    "        return up,right\n",
    "    \n",
    "    def get_down_left(self,i,j):\n",
    "        '''\n",
    "        PBCs down/left\n",
    "        '''\n",
    "        if j==0:\n",
    "            down = self.L-1\n",
    "        else:\n",
    "            down = j-1\n",
    "        if i==0:\n",
    "            left = self.L-1\n",
    "        else:\n",
    "            left = i-1\n",
    "            \n",
    "        return down,left\n",
    "    \n",
    "    def calc_energy(self):\n",
    "        '''\n",
    "        Calculates the total energy.\n",
    "        NB: this is used only at initialization.\n",
    "        '''\n",
    "        E = 0\n",
    "        for j in range(self.L):\n",
    "            for i in range(self.L):\n",
    "                up,right = self.get_up_right(i,j)\n",
    "                E = E - self.spin[i,j]*(self.spin[i,up]+self.spin[right,j])\n",
    "        return E\n",
    "                       \n",
    "    def deltaE(self,i,j):\n",
    "        '''\n",
    "        Calculates change in energy for the spin flip\n",
    "        '''\n",
    "        up,right = self.get_up_right(i,j)\n",
    "        down,left = self.get_down_left(i,j)\n",
    "        res = 2.0*self.spin[i,j]*(self.spin[i,up]+self.spin[i,down]+self.spin[left,j]+self.spin[right,j])\n",
    "        return (res,int(res))\n",
    "    \n",
    "    def initialize(self):\n",
    "        '''\n",
    "        Initialization of the system:\n",
    "        -spins are randomly configured\n",
    "        -energy is calculated from scratch\n",
    "        -dictionary with Metropolis' probability\n",
    "        coefficents is created\n",
    "        '''\n",
    "        r = self.rng.random((self.L,self.L))\n",
    "        self.spin = self.make_pm(r)\n",
    "        self.M = np.sum(self.spin)\n",
    "        self.M_N = self.M/float(self.N)\n",
    "        self.E = self.calc_energy()\n",
    "        self.E_N = self.E/float(self.N)\n",
    "        for dE in range(-8,8+1,4):\n",
    "            self.w_dict[dE] = np.exp(-1.0*dE/self.T)\n",
    "    \n",
    "    def metro_step(self):\n",
    "        'One step of Metropolis spin flip dynamics'\n",
    "        for ispin in range(self.N):\n",
    "            (ii,jj) = self.rng.integers(0,high=self.L,size=2)\n",
    "            dE,dE_int = self.deltaE(ii,jj)\n",
    "            rnd_w = self.rng.random()\n",
    "            if (rnd_w<=self.w_dict[dE_int]):\n",
    "                self.spin[ii,jj] = -1*self.spin[ii,jj]\n",
    "                self.acc += 1\n",
    "                self.M = self.M + 2* self.spin[ii,jj]\n",
    "                self.M_N = self.M/float(self.N)\n",
    "                self.E = self.E + dE\n",
    "                self.E_N = self.E/float(self.N)\n",
    "    \n",
    "\n",
    "    def metro_driver(self,N_step,store=True):\n",
    "        'Run N Metropolis steps of spin flip dynamics'\n",
    "        for i in range(N_step):\n",
    "            self.counter += self.N\n",
    "            self.metro_step()\n",
    "            if store:\n",
    "                self.data.append((self.E,self.M))            \n",
    "            \n",
    "    def ising_driver(self):\n",
    "        '''\n",
    "        Main driver for the Monte Carlo\n",
    "        '''\n",
    "        self.initialize()\n",
    "        self.metro_driver(self.N_equi,store=self.sample_equi)\n",
    "        self.metro_driver(self.N_mcs)\n",
    "        self.data = np.array(self.data)\n",
    "        np.savetxt(self.data_filename,self.data)\n",
    "    \n",
    "    @staticmethod\n",
    "    def plot_spin(spin):\n",
    "        x,y = np.meshgrid(range(spin.shape[0]),range(spin.shape[1]))\n",
    "        plt.xlabel('x')\n",
    "        plt.ylabel('y')\n",
    "        col_dict = {-1: \"orange\",\n",
    "            1: \"green\"}\n",
    "        cmap = ListedColormap([col_dict[x] for x in col_dict.keys()])\n",
    "        s = plt.scatter(x,y,c=spin[x,y],cmap=cmap,label='spin')\n",
    "        plt.title('spin configuration')\n",
    "        cax = plt.colorbar(s, ticks=[-1,1])\n",
    "        plt.show()\n",
    "    \n",
    "    @staticmethod\n",
    "    def calc_autocorr(a,N):\n",
    "        'Autocorrelation function'\n",
    "        res = np.ones(N)\n",
    "        for k in range(1,N):\n",
    "            c_k = np.average(a[:-k]*a[k:]) \n",
    "            x_r = np.average(a[:-k])\n",
    "            x_r_2 = np.average(a[:-k]**2)\n",
    "            res[k] = (c_k-x_r**2)/(x_r_2-x_r**2)\n",
    "        return res\n",
    "    \n",
    "    def get_stats(self):\n",
    "        '''\n",
    "        Calculate, store and print \n",
    "        statistical quantities\n",
    "        '''\n",
    "        self.ar = self.acc*1.0/self.counter\n",
    "        self.E_N_av =np.average(self.data[:,0])/self.N\n",
    "        self.M_N_av = np.average(self.data[:,1])/self.N\n",
    "        self.c_heat = (np.average(self.data[:,0]**2)- (self.E_N_av*self.N)**2)/(self.T**2*self.N)\n",
    "        self.chi_mag = (np.average(self.data[:,1]**2)- (self.M_N_av*self.N)**2)/(self.T**2*self.N)\n",
    "        print(\"Total number of steps run = {}\".format(self.counter/self.N))\n",
    "        print(\"Acceptance ratio = {}\".format(self.ar))\n",
    "        print(\"Temperature = {}\".format(self.T))\n",
    "        print(\"<E/N> = {}\".format(self.E_N_av))\n",
    "        print(\"<M/N> = {}\".format(self.M_N_av))\n",
    "        print(\"specific heat = {}\".format(self.c_heat))\n",
    "        print(\"magnetic susceptibility = {}\".format(self.chi_mag))\n",
    "        \n",
    "    #def get_autocorr(self,n_autocorr=100):\n",
    "    #    '''\n",
    "    #    Wrapper for the autocorrelation\n",
    "    #    function calculator\n",
    "    #    '''\n",
    "    #    a = np.array(self.data)\n",
    "    #    self.autocorr = self.calc_autocorr(a,n_autocorr)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "aa0e820f",
   "metadata": {},
   "source": [
    "Let's run the code!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fc5c06a0",
   "metadata": {},
   "outputs": [],
   "source": [
    "# lateral size L (N=LxL), temperature T, nequil, nmcas\n",
    "ising = Ising(30,2,1000,1000,sample_equi=True)\n",
    "ising.initialize()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9257ef94",
   "metadata": {},
   "outputs": [],
   "source": [
    "#array with current spin configuration\n",
    "ising.spin"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0f895254",
   "metadata": {},
   "outputs": [],
   "source": [
    "ising.plot_spin(ising.spin)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4925510b",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Running Monte Carlo\n",
    "ising.ising_driver()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f6784852",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Extracting key quantities\n",
    "ising.get_stats()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ecf12a11",
   "metadata": {},
   "outputs": [],
   "source": [
    "ising.spin"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a03ecaac",
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.ylabel('Magnetisation')\n",
    "plt.xlabel('MC step')\n",
    "plt.plot(np.abs(ising.data[:,1]/ising.N),'-',label='|M|')\n",
    "plt.legend()\n",
    "plt.show()\n",
    "plt.close()\n",
    "plt.ylabel('Energy')\n",
    "plt.xlabel('MC step')\n",
    "plt.plot(ising.data[:,0]/ising.N,'-',label='E',c='orange')\n",
    "plt.legend()\n",
    "plt.show()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "89971652",
   "metadata": {},
   "outputs": [],
   "source": [
    "ising.plot_spin(ising.spin)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e9cb1f83",
   "metadata": {},
   "source": [
    "-----------------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "cd076f02",
   "metadata": {},
   "source": [
    "Now we increase the temperature to T=4."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c2d15d9c",
   "metadata": {},
   "outputs": [],
   "source": [
    "# lateral size L (N=LxL), temperature T, nequil, nmcas\n",
    "ising = Ising(30,4,1000,1000,sample_equi=True)\n",
    "ising.initialize()\n",
    "ising.ising_driver()\n",
    "ising.get_stats()\n",
    "ising.plot_spin(ising.spin)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "99f03430",
   "metadata": {},
   "source": [
    "-------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9ae7729b",
   "metadata": {},
   "source": [
    ">There seems to be a phase transition between T=2 and T=4, let's investigate what happens as a function of temperature."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9e2ebc2f",
   "metadata": {},
   "outputs": [],
   "source": [
    "res=[]\n",
    "T_list = np.linspace(1.5,4,10)\n",
    "for T in T_list:\n",
    "    ising = Ising(30,T,1000,1000,sample_equi=False)\n",
    "    ising.initialize()\n",
    "    ising.ising_driver()\n",
    "    ising.get_stats()\n",
    "    res.append((ising.E_N_av,ising.M_N_av,ising.c_heat,ising.chi_mag))\n",
    "res = np.array(res)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f75640b1",
   "metadata": {},
   "outputs": [],
   "source": [
    "#np.savetxt('res_acc.txt',res)\n",
    "#np.savetxt('T_acc.txt',T_list)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ddb12ee0",
   "metadata": {},
   "outputs": [],
   "source": [
    "#plotting as a function of temperature\n",
    "def plot_res(T_list,res):\n",
    "    fig, axs = plt.subplots(ncols=2, nrows=2,layout=\"constrained\")\n",
    "    axs[0,0].plot(T_list,res[:,0],'o-',c='orange',label='energy')\n",
    "    axs[0,1].plot(T_list,np.abs(res[:,1]),'o-',label='magnetisation')\n",
    "    axs[1,0].plot(T_list,res[:,2],'o-',c='red',label='specific heat')\n",
    "    axs[1,1].plot(T_list,res[:,3],'o-',c='k',label='magnetic susceptibility')\n",
    "    for axx in axs:\n",
    "        for ax in axx:\n",
    "            ax.legend()\n",
    "            ax.set_xlabel('T')\n",
    "            ax.set_xlim([1.5,4])\n",
    "            ax.axvline(x=2.26,c='grey')\n",
    "plot_res(T_list,res)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7615d6f2",
   "metadata": {},
   "source": [
    ">A proper sampling requires more computational time, let's load some results computed with this code for longer runs (10000 steps) and with a more dense sampling of temperatures."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "affceff0",
   "metadata": {},
   "outputs": [],
   "source": [
    "res = np.loadtxt('./res_acc.txt')\n",
    "T_list = np.loadtxt('./T_acc.txt')\n",
    "plot_res(T_list,res)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "21b46b42",
   "metadata": {},
   "source": [
    "### Profiling\n",
    "In the next lecture we will see how to make the code faster by integrating Fortran subroutines into Python code with *f2py*.\\\n",
    "For the moment, we will just investigate where the code spends most of the run time."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "fa4bca4d",
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext line_profiler"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "1964bcb2",
   "metadata": {},
   "outputs": [],
   "source": [
    "ising = Ising(10,T,100,1000,sample_equi=False)\n",
    "ising.initialize()\n",
    "%lprun -f ising.ising_driver ising.ising_driver()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "7863e612",
   "metadata": {},
   "outputs": [],
   "source": [
    "ising = Ising(10,T,100,1000,sample_equi=False)\n",
    "ising.initialize()\n",
    "%lprun -f ising.metro_step ising.ising_driver()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "-----------------  \n",
    "\n",
    "# f2py - interfacing Fortran code to Python\n",
    "\n",
    "From https://numpy.org/doc/stable/f2py/ :\n",
    "\n",
    "\"The purpose of the F2PY –Fortran to Python interface generator– utility is to provide a connection between Python and Fortran. F2PY is a part of NumPy (numpy.f2py) and also available as a standalone command line tool.\n",
    "\n",
    "F2PY facilitates creating/building Python C/API extension modules that make it possible\n",
    "\n",
    "* to call Fortran 77/90/95 external subroutines and Fortran 90/95 module subroutines as well as C functions;\n",
    "* to access Fortran 77 COMMON blocks and Fortran 90/95 module data, including allocatable arrays\n",
    "\n",
    "from Python\".\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "There are different ways to use f2py, today we will first create an interface and then compile the Fortran code as a Python module:\n",
    "\n",
    "\n",
    "* python3 -m numpy.f2py isingmetro.f90 -m isingf -h ising.pyf\n",
    "* python3 -m numpy.f2py -c ising.pyf isingmetro.f90"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!rm ising.pyf\n",
    "!python3 -m numpy.f2py isingmetro.f90 -m isingf -h ising.pyf"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "!python3 -m numpy.f2py -c ising.pyf isingmetro.f90"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "ls"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "# useful function recycled from the Python class for the Ising model\n",
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "def get_up_right(L,i,j):\n",
    "        '''\n",
    "        PBCs up/right\n",
    "        '''\n",
    "        if j==(L-1):\n",
    "            up = 0\n",
    "        else:\n",
    "            up = j+1\n",
    "        if i==(L-1):\n",
    "            right = 0\n",
    "        else:\n",
    "            right = i+1\n",
    "            \n",
    "        return up,right\n",
    "def get_down_left(L,i,j):\n",
    "        '''\n",
    "        PBCs down/left\n",
    "        '''\n",
    "        if j==0:\n",
    "            down = L-1\n",
    "        else:\n",
    "            down = j-1\n",
    "        if i==0:\n",
    "            left = L-1\n",
    "        else:\n",
    "            left = i-1\n",
    "            \n",
    "        return down,left\n",
    "def calc_energy(L,spin):\n",
    "        '''\n",
    "        Calculates the total energy.\n",
    "        NB: this is used only at initialization.\n",
    "        '''\n",
    "        E = 0\n",
    "        for j in range(L):\n",
    "            for i in range(L):\n",
    "                up,right = get_up_right(L,i,j)\n",
    "                E = E - spin[i,j]*(spin[i,up]+spin[right,j])\n",
    "        return E\n",
    "def get_stats(data,N,T):\n",
    "        '''\n",
    "        Calculate, store and print \n",
    "        statistical quantities\n",
    "        '''\n",
    "        E_N_av =np.average(data[:,0])/N\n",
    "        M_N_av = np.average(data[:,1])/N\n",
    "        c_heat = (np.average(data[:,0]**2)- (E_N_av*N)**2)/(T**2*N)\n",
    "        chi_mag = (np.average(data[:,1]**2)- (M_N_av*N)**2)/(T**2*N)\n",
    "        #print(\"Total number of steps run = {}\".format(counter/N))\n",
    "        #print(\"Acceptance ratio = {}\".format(ar))\n",
    "        #print(\"Temperature = {}\".format(self.T))\n",
    "        print(\"<E/N> = {}\".format(E_N_av))\n",
    "        print(\"<M/N> = {}\".format(M_N_av))\n",
    "        print(\"specific heat = {}\".format(c_heat))\n",
    "        print(\"magnetic susceptibility = {}\".format(chi_mag))     \n",
    "        return E_N_av,M_N_av,c_heat,chi_mag"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from isingf import isingf"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.t  = 1\n",
    "isingf.l = 30\n",
    "isingf.n = 30*30"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "#isingf.alloc()\n",
    "isingf.init()\n",
    "isingf.w"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.t  = 1\n",
    "isingf.l = 30\n",
    "isingf.n = 30*30"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "#isingf.alloc()\n",
    "isingf.init()\n",
    "isingf.w"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.spin = np.ones((30,30),dtype=np.int32,order='F')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.e = calc_energy(isingf.l,isingf.spin)\n",
    "isingf.m = np.sum(isingf.spin)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.e,isingf.m"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.metropolis()\n",
    "isingf.spin"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.e"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "N_eq = 10000\n",
    "data=np.zeros((N_eq,2))\n",
    "for i in range(N_eq):\n",
    "    isingf.metropolis()\n",
    "    data[i,:] = (isingf.e,isingf.m)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.xlabel('step')\n",
    "plt.ylabel('magnetization')\n",
    "plt.plot(np.abs(data[:,1])/isingf.n,'o-')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "get_stats(data,isingf.n,isingf.t)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from isingf import isingf\n",
    "N_eq = 1000\n",
    "N_mc = 100000\n",
    "T_list = np.linspace(1.5,4,30)\n",
    "res_f = []\n",
    "for T in T_list:\n",
    "    isingf.t  = T\n",
    "    isingf.l = 30\n",
    "    isingf.n = 30*30\n",
    "    #isingf.alloc()\n",
    "    isingf.init()\n",
    "    isingf.spin = np.ones((30,30),dtype=np.int32,order='F')\n",
    "    isingf.e = calc_energy(isingf.l,isingf.spin)\n",
    "    isingf.m = np.sum(isingf.spin)\n",
    "    data=np.zeros((N_mc,2))\n",
    "    for i in range(N_eq):\n",
    "        isingf.metropolis()\n",
    "    for i in range(N_mc):\n",
    "        isingf.metropolis()\n",
    "        data[i,:] = (isingf.e,isingf.m)\n",
    "    e,m,heat,chi = get_stats(data,isingf.n,isingf.t)\n",
    "    res_f.append([e,m,heat,chi])\n",
    "#    del isingf\n",
    "res_f = np.array(res_f)\n",
    "\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "plot_res(T_list,res_f)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "> Notice how much faster this code is with respect to the pure Python code!"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Profiling - revisited"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "from isingf import isingf\n",
    "def run_isingf(T,N_eq,N_mc):\n",
    "    isingf.t  = T\n",
    "    isingf.l = 30\n",
    "    isingf.n = 30*30\n",
    "    #isingf.alloc()\n",
    "    isingf.init()\n",
    "    isingf.spin = np.ones((30,30),dtype=np.int32,order='F')\n",
    "    isingf.e = calc_energy(isingf.l,isingf.spin)\n",
    "    isingf.m = np.sum(isingf.spin)\n",
    "    data=np.zeros((N_mc,2))\n",
    "    for i in range(N_eq):\n",
    "        isingf.metropolis()\n",
    "    for i in range(N_mc):\n",
    "        isingf.metropolis()\n",
    "        data[i,:] = (isingf.e,isingf.m)\n",
    "    return get_stats(data,isingf.n,isingf.t)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "ising = Ising(30,2,1000,2000,sample_equi=False)\n",
    "ising.initialize()\n",
    "%time ising.ising_driver() \n",
    "%time run_isingf(2,1000,2000) "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext line_profiler\n",
    "%lprun -f run_isingf run_isingf(2,1000,2000) "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "isingf.w"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "fis_comp_2023",
   "language": "python",
   "name": "fis_comp_2023"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
